package nondeterministic_fixture_test

import (
	. "github.com/onsi/ginkgo/v2"
)

var specGenerator = map[string]bool{
	"map-A": true,
	"map-B": true,
	"map-C": true,
	"map-D": true,
	"map-E": true,
	"map-F": true,
	"map-G": true,
	"map-H": true,
	"map-I": true,
	"map-J": true,
}

var _ = Describe("some tests, that include a test generated by iterating over a map", func() {
	Describe("some tests", func() {
		It("runs A", func() {

		})
		It("runs B", func() {

		})
	})

	When("iterating over a map", func() {
		for key := range specGenerator {
			It("runs "+key, func() {

			})
		}
	})

	It("runs some other tests", func() {

	})
})
